import configparser
import sys
import threading
import time
from json import JSONDecodeError

import pika
import logging
import json

from pika.exceptions import ConnectionClosedByBroker, AMQPChannelError, AMQPConnectionError

logging.basicConfig(format=u'%(filename)s[LINE:%(lineno)d]# %(levelname)-8s [%(asctime)s]  %(message)s',
                    level=logging.DEBUG)
logging.getLogger('pika').setLevel(logging.WARNING)


class PikaConsumer:

    reconnection_timeout = 5

    def __init__(self, handlers, queue, host, port, user, password):
        self._queue = queue
        self._host = host
        self._port = port
        self._user = user
        self._pass = password
        self._terminate = False
        self._handlers = handlers
        self._connection = None
        self._channel = None
        self._disconnection_cs = threading.Lock()
        self._active_callbacks = 0
        self._active_threads = []
        # Here we will store messages, which are really processed, but the acknowledge wasn't sent
        self._repeatable_messages = []

    @classmethod
    def create_from_config(cls, handlers, config_path, config_section='consumer'):
        config = configparser.ConfigParser()
        config.readfp(open(config_path))
        queue = config.get(config_section, "queue")
        host = config.get(config_section, "host")
        port = config.get(config_section, "port")
        user = config.get(config_section, "user")
        password = config.get(config_section, "password")
        return cls(handlers, queue, host, port, user, password)

    def __del__(self):
        self.terminate()

    def terminate(self):
        self._terminate = True
        self.disconnect()

    def connect(self):
        logging.info("start connecting on host %s queue %s" % (self._host, self._queue))
        credentials = pika.PlainCredentials(self._user, self._pass)
        parameters = pika.ConnectionParameters(host=self._host, port=self._port, credentials=credentials)
        self._connection = pika.BlockingConnection(parameters)
        self._channel = self._connection.channel()
        self._channel.basic_qos(prefetch_count=1)
        self._channel.queue_declare(queue=self._queue, durable=True)
        self._channel.basic_consume(queue=self._queue, on_message_callback=self._callback)

    def disconnect(self):
        self._disconnection_cs.acquire()
        if self._connection is not None and not self._connection.is_closed:
            logging.info("start server disconnection on host %s queue %s" % (self._host, self._queue))
            try:
                self._channel.stop_consuming()
            except Exception as ex:
                logging.error("rabbit stop consuming error: %s" % str(ex),
                              extra={"queue": self._queue, "host": self._host})
            try:
                while self._active_callbacks > 0:
                    time.sleep(1)
                self._connection.close()
            except Exception as ex:
                logging.error("rabbit disconnect error: %s" % str(ex),
                              extra={"queue": self._queue, "host": self._host})
        self._disconnection_cs.release()

    def start(self):
        while not self._terminate:
            try:
                self.connect()
                logging.info("start consuming on host %s queue %s" % (self._host, self._queue))
                self._channel.start_consuming()
            except ConnectionClosedByBroker as ex:
                if not self._terminate:
                    logging.error("server broke the connection: %s in line %s, reconnecting in %d seconds..." %
                              (str(sys.exc_info()[-1].tb_lineno), str(ex), self.reconnection_timeout),
                              extra={"queue": self._queue, "host": self._host})
                    time.sleep(self.reconnection_timeout)
            except AMQPChannelError as ex:
                if not self._terminate:
                    logging.error("caught a channel error in line %s: %s, reconnecting..." %
                              (str(sys.exc_info()[-1].tb_lineno), str(ex)),
                              extra={"queue": self._queue, "host": self._host})
            except AMQPConnectionError as ex:
                if not self._terminate:
                    logging.error("connection was closed in line %s: %s, reconnecting in %d seconds..." %
                              (str(sys.exc_info()[-1].tb_lineno), str(ex), self.reconnection_timeout),
                              extra={"queue": self._queue, "host": self._host})
                    time.sleep(self.reconnection_timeout)
            except Exception as ex:
                if not self._terminate:
                    logging.error("start error in line %s: %s, restart in %d seconds" %
                                  (str(sys.exc_info()[-1].tb_lineno), str(ex), self.reconnection_timeout),
                                  extra={"queue": self._queue, "host": self._host})
                    time.sleep(self.reconnection_timeout)
            finally:
                self.disconnect()
        logging.info("rabbit consumer terminated on host %s queue %s" % (self._host, self._queue))

    def _callback(self, ch, method, properties, body):
        try:
            self._active_callbacks += 1
            message = json.loads(body.decode("utf-8"))
            if "action" in message.keys():
                action = message["action"]
                unique = message["rmq_unique"]
                if action in self._handlers.keys():
                    if unique not in self._repeatable_messages:
                        if self._handlers[action](message):
                            try:
                                ch.basic_ack(delivery_tag=method.delivery_tag)
                            except Exception as ex:
                                logging.error("cannot send acknowledge: %s" % str(ex))
                                self._repeatable_messages.append(unique)
                    else:
                        ch.basic_ack(delivery_tag=method.delivery_tag)
                else:
                    logging.error("no such handler: %s" % action)
            else:
                logging.error("wrong message format: no column action")
        except JSONDecodeError as ex:
            logging.error("json decode error in line %s: %s" % (str(sys.exc_info()[-1].tb_lineno), str(ex)))
        except Exception as ex:
            logging.error("rabbit callback error in line %s: %s" % (str(sys.exc_info()[-1].tb_lineno), str(ex)),
                          extra={"queue": self._queue, "body": body, "host": self._host})
        finally:
            self._active_callbacks -= 1




import configparser
import sys
import time
import uuid

import pika
import logging
import json

from pika.exceptions import NackError

logging.basicConfig(format=u'%(filename)s[LINE:%(lineno)d]# %(levelname)-8s [%(asctime)s]  %(message)s',
                    level=logging.DEBUG)
logging.getLogger('pika').setLevel(logging.WARNING)


class PikaPublisher:

    reconnection_timeout = 5
    retry_timeout = 1

    # WARNING pika is not thread safe, so use one sender instance for one thread
    def __init__(self, queue, host, port, user, password, max_length=None):
        self._queue = queue
        self._host = host
        self._port = port
        self._user = user
        self._pass = password
        self._connection = None
        self._channel = None
        self._max_length = max_length
        self.connected = False
        self.connect()

    @classmethod
    def create_from_config(cls, config_path, config_section='publisher'):
        config = configparser.ConfigParser()
        config.readfp(open(config_path))
        queue = config.get(config_section, "queue")
        host = config.get(config_section, "host")
        port = config.get(config_section, "port")
        user = config.get(config_section, "user")
        password = config.get(config_section, "password")
        max_length = config.get(config_section, "max_length", fallback=None)
        return cls(queue, host, port, user, password, max_length)

    def connect(self, timeout=1000.0):
        start_time = time.time()
        while not self.connected:
            try:
                credentials = pika.PlainCredentials(self._user, self._pass)
                parameters = pika.ConnectionParameters(host=self._host, port=self._port, credentials=credentials)
                self._connection = pika.BlockingConnection(parameters)
                self._channel = self._connection.channel()
                arguments = None
                if self._max_length is not None:
                    arguments = {"x-max-length": int(self._max_length), "x-overflow": "reject-publish"}
                self._channel.queue_declare(queue=self._queue, durable=True, arguments=arguments)
                self._channel.confirm_delivery()
                self.connected = True
            except Exception as ex:
                if time.time()-start_time > timeout:
                    logging.error("connecting error: %s" % (str(ex)),
                                  extra={"queue": self._queue, "host": self._host})
                    raise ex
                logging.error("connecting error: %s, reconnecting in %d seconds..." %
                              (str(ex), self.reconnection_timeout),
                              extra={"queue": self._queue, "host": self._host})
                time.sleep(self.reconnection_timeout)

    def disconnect(self):
        self.connected = False
        if self._connection is not None and not self._connection.is_closed:
            try:
                self._connection.close()
            except Exception as ex:
                logging.error("rabbit disconnect error: %s" % str(ex),
                              extra={"queue": self._queue, "host": self._host})

    def publish(self, message, timeout=1000.0):
        try:
            start_time = time.time()
            if "rmq_unique" not in message.keys():
                message["rmq_unique"] = str(uuid.uuid4())
            body = json.dumps(message)
            sent = False
            while not sent:
                try:
                    self._channel.basic_publish(exchange='', routing_key=self._queue, body=body,
                                            properties=pika.BasicProperties(
                                                 delivery_mode=2,
                                            ), mandatory=True)
                    sent = True
                except NackError as ex:
                    logging.error("queue %s is overflowed, retry in %d seconds..." % (self._queue, self.retry_timeout),
                                  extra={"queue": self._queue, "host": self._host})
                    time.sleep(self.retry_timeout)
                except Exception as ex:
                    if time.time() - start_time > timeout:
                        logging.error("publish error: %s" % (str(ex)),
                                      extra={"queue": self._queue, "host": self._host})
                        self.disconnect()
                        raise ex
                    logging.error("publish error: %s, reconnecting in %d seconds..." %
                                  (str(ex), self.reconnection_timeout),
                                  extra={"queue": self._queue, "host": self._host})
                    self.disconnect()
                    time.sleep(self.reconnection_timeout)
                    self.connect(timeout)
        except Exception as ex:
            logging.error("rabbit publish error in line %s: %s" % (str(sys.exc_info()[-1].tb_lineno), str(ex)),
                          extra={"queue": self._queue, "message": message})
            raise ex

    def __del__(self):
        self.disconnect()



import os
import random
import string
import threading
import time

from pikajson.client import PikaPublisher
from pikajson.server import PikaConsumer


class PerformanceTest:

    def __init__(self, wait_time=0.1, action='test', messages_count=100, clients_count=10):
        self.terminate = False
        self.server = None
        self.server_thread = None
        self.client_threads = []
        self.clients = []
        self.clients_timings = []
        self.wait_time = wait_time
        self.action = action
        self.messages_count = messages_count
        self.clients_count = clients_count
        self.server_messages = []
        self.clients_messages = {}
        self.server_work_time = 0
        self.server_termination_time = 0
        self.server_start_time = time.time()

    def test_server(self, message):
        self.server_messages.append(message)
        return True

    def test_waiting(self, message):
        time.sleep(self.wait_time)
        self.server_messages.append(message)
        return True

    def test_threading(self, message):
        thread = threading.Thread(target=self.test_waiting, args=(message,))
        thread.daemon = True
        thread.start()
        return True

    def create_server(self):
        path = os.path.dirname(os.path.abspath(__file__))
        self.server = PikaConsumer.create_from_config({"test": self.test_server,
                                                       "test_waiting": self.test_waiting,
                                                       "test_threading": self.test_threading},
                                                       path + '/configs/consumer.ini')

    def start_server(self):
        self.server_start_time = time.time()
        if self.server is not None:
            self.server_thread = threading.Thread(target=self.server.start)
            self.server_thread.daemon = True
            self.server_thread.start()

    def stop_server(self):
        if self.server is not None and self.server_thread is not None:
            start_time = time.time()
            self.server.terminate()
            self.server_thread.join()
            self.server_work_time = time.time() - self.server_start_time
            self.server_termination_time = time.time() - start_time

    def client_sending(self, number):
        self.clients_messages[number] = []
        path = os.path.dirname(os.path.abspath(__file__))
        client = PikaPublisher.create_from_config(path + '/configs/publisher.ini')
        self.clients.append(client)
        length = random.randint(10, 1000)
        message = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(length))
        i = 0
        while i < self.messages_count:
            i += 1
            start_time = time.time()
            client.publish({"action": self.action, "test": message})
            self.clients_timings.append(time.time()-start_time)
            self.clients_messages[number].append(message)

    def start_clients(self):
        i = 0
        while i < self.clients_count:
            i += 1
            client_thread = threading.Thread(target=self.client_sending, args=(i, ))
            client_thread.daemon = True
            client_thread.start()
            self.client_threads.append(client_thread)

    def wait_for_clients(self):
        for thread in self.client_threads:
            thread.join()
